# -*- Mode: Python -*-

cdef extern from "loess.h":
    ctypedef struct c_loess_errstatus "loess_errstatus":
        int err_status
        char *err_msg
    ctypedef struct c_loess_inputs "loess_inputs":
        long   n
        long   p
        double *y
        double *x
        double *weights
    ctypedef struct c_loess_model "loess_model":
        double span
        int    degree
        int    normalize
        int    parametric[8]
        int    drop_square[8]
        char   *family
    ctypedef struct c_loess_control "loess_control":
        char   *surface
        char   *statistics
        double cell
        char   *trace_hat
        int    iterations
    ctypedef struct c_loess_kd_tree "loess_kd_tree":
        pass
    ctypedef struct c_loess_outputs "loess_outputs":
        double  *fitted_values
        double  *fitted_residuals
        double  enp
        double  residual_scale
        double  one_delta
        double  two_delta
        double  *pseudovalues
        double  trace_hat
        double  *diagonal
        double  *robust
        double  *divisor
    ctypedef struct c_loess "loess":
        c_loess_inputs *inputs
        c_loess_model *model
        c_loess_control *control
        c_loess_kd_tree *kd_tree
        c_loess_outputs *outputs
        c_loess_errstatus status
    ctypedef struct c_prediction "prediction":
        double  *fit
        double  *se_fit
        int se
        int m
        double  residual_scale
        double  df
#    ctypedef struct c_anova "anova_struct":
#        double  dfn
#        double  dfd
#        double  F_value
#        double  Pr_F
    ctypedef struct c_confidence_intervals "confidence_intervals":
        double  *fit
        double  *upper
        double  *lower

    void loess_setup(double *x, double *y, double *w, long n, long p, c_loess *lo)
    void loess_model_setup(c_loess_model *model)
    void loess_inputs_setup(double *x, double *y, double *w, long n, long p, c_loess_inputs *inputs)
    void loess_outputs_setup(long n, long p, c_loess_outputs *outputs)
    void loess_kd_tree_setup(long n, long p, c_loess_kd_tree *kd_tree)
    void loess_control_setup(c_loess_control *control)

    void loess_fit(c_loess *lo)

    void loess_inputs_free(c_loess_inputs *inputs)
    void loess_outputs_free(c_loess_outputs *outputs)
    void loess_kd_tree_free(c_loess_kd_tree *kd_tree)
    void loess_free_mem(c_loess *lo)

    void loess_summary(c_loess *lo)

    void predict_setup(c_prediction *pre, c_loess *lo, int se, int m);
    void c_predict "predict" (double *eval, c_loess *lo, c_prediction *pre)
    void predict_free(c_prediction *pre)

    void c_pointwise "pointwise" (c_prediction *pre, double coverage, c_confidence_intervals *ci)
    double pf (double q, double df1, double df2)
    double ibeta (double x, double a, double b)
    void pw_free_mem (c_confidence_intervals *ci)



#!/usr/bin/env python
from __future__ import annotations

import os
import re
import shlex
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Sequence

DEFAULT_DESCRIBE = [
    "git",
    "describe",
    "--dirty",
    "--tags",
    "--long",
    "--match",
    "*[0-9]*",
]

LAST_COMMIT_SHORTHASH = [
    "git",
    "log",
    "--format='%h'",
    "-n",
    "1",
]

count = r"(?:[0-9]|[1-9][0-9]+)"
DESCRIBE_PATTERN = re.compile(
    r"^v"
    rf"(?P<version>{count}\.{count}\.{count})"
    rf"(?P<pre>(a|b|rc|alpha|beta){count})?"
    r"(-(?P<commits>\d+)-g(?P<hash>[a-z0-9]+))?"
    r"(?P<dirty>-dirty)?"
    r"$"
)

VERSION_LINE_PATTERN = re.compile(
    r'__version__ = "(?P<version>.+?)"'
)

NULL_VERSION = "0.0.0"


def run(cmd: str | Sequence[str]) -> str:
    if isinstance(cmd, str) and os.name == "posix":
        cmd = shlex.split(cmd)
    with Popen(
        cmd,
        stdin=PIPE,
        stderr=PIPE,
        stdout=PIPE,
        text=True,
        encoding="utf-8"
    ) as p:
        stdout, _ = p.communicate()
    return stdout.strip()


def is_git() -> bool:
    """
    Return True if inside a git repo
    """
    res = run("git rev-parse --is-inside-work-tree")
    return res == "return"


def get_version_from_scm() -> str:
    """
    Return a SemVer (& PEP440 compliant) from `git describe` output
    """
    desc = run(DEFAULT_DESCRIBE)
    m = DESCRIBE_PATTERN.match(desc)

    if not m:
        return ""

    pre = m.group("pre") or ""
    version = m.group("version")
    commits = m.group("commits")

    if commits != "0":
        v = f"{version}.dev{commits}"
    else:
        v = f"{version}{pre}"

    return v


def get_version_from_file() -> str:
    """
    Read version from package/_version.py
    """
    file = Path(__file__).parent.parent / "_version.py"

    if file.exists():
        s = file.read_text()
        match = VERSION_LINE_PATTERN.search(s)
        if match:
            return match.group('version')
    return ""


def get_version() -> str:
    """
    Return a SemVer (& PEP440 compliant) from `git describe` output
    """
    return get_version_from_scm() or get_version_from_file() or NULL_VERSION


if __name__ == "__main__":
    print(get_version())



""" Distributor init file

Distributors: you can add custom code here to support particular distributions
of scikit-misc.

For example, this is a good place to put any checks for hardware requirements.

The scikit-misc standard source distribution will not put code in this file, so
you can safely replace this file with your own version.
"""



#!python
""" Detect bitness (32 or 64) of Mingw-w64 gcc build target on Windows.
"""
# Ref: https://github.com/scipy/scipy/blob/main/scipy/_build_utils/gcc_build_bitness.py  # noqa: E501
# commit: 6c4a3f3

import re
from subprocess import run


def main():
    res = run(
        ['gcc', '-v'],
        check=True,
        text=True,
        capture_output=True
    )
    target = re.search(
        r'^Target: (.*)$',
        res.stderr,
        flags=re.M
    ).groups()[0]

    if target.startswith('i686'):
        print('32')
    elif target.startswith('x86_64'):
        print('64')
    else:
        raise RuntimeError("Could not detect Mingw-w64 bitness")


if __name__ == "__main__":
    main()



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version('scikit_misc')
except PackageNotFoundError:
    # package is not installed
    pass
finally:
    del version
    del PackageNotFoundError

__all__ = ['__version__']

try:
    from skmisc.__config__ import show as show_config  # noqa: F401
except ImportError as err:
    msg = """Error importing skmisc: you cannot import skmisc while
    being in skmisc source directory; please exit the skmisc source
    tree first, and relaunch your python intepreter."""
    raise ImportError('\n\n'.join([err.message, msg]))  # type: ignore
else:
    __all__.append('show_config')


    def test(args=None, plugins=None):
        """
        Run tests
        """
        from pathlib import Path
        # The doctests are not run when called from an installed
        # package since the pytest.ini is not included in the
        # package.
        try:
            import pytest
        except ImportError:
            msg = "To run the tests, you must install pytest"
            raise ImportError(msg)
        path = str(Path(__file__).parent)
        if args is None:
            args = [path]
        else:
            args.append(path)
        return pytest.main(args=args, plugins=plugins)



import pickle
import tempfile
from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

from skmisc.loess import loess, loess_anova

data_path = Path(__file__).parent


def madeup_data():
    dfile = data_path / 'madeup_data'
    rfile = data_path / 'madeup_result'

    with open(dfile, 'r') as f:
        f.readline()
        x = np.fromiter(
            (float(v) for v in f.readline().rstrip().split()),
            np.float_).reshape(-1, 2)
        f.readline()
        y = np.fromiter(
            (float(v) for v in f.readline().rstrip().split()),
            np.float_)

    results = []
    with open(rfile, 'r') as f:
        for i in range(8):
            f.readline()
            z = np.fromiter(
                (float(v) for v in f.readline().rstrip().split()),
                np.float_)
            results.append(z)

    newdata1 = np.array([[-2.5, 0.], [2.5, 0.], [0., 0.]])
    newdata2 = np.array([[-0.5, 0.5], [0., 0.]])
    return (x, y, results, newdata1, newdata2)


def gas_data():
    NOx = np.array([4.818, 2.849, 3.275, 4.691, 4.255, 5.064, 2.118, 4.602,
                    2.286, 0.970, 3.965, 5.344, 3.834, 1.990, 5.199, 5.283,
                    3.752, 0.537, 1.640, 5.055, 4.937, 1.561])
    E = np.array([0.831, 1.045, 1.021, 0.970, 0.825, 0.891, 0.71, 0.801,
                  1.074, 1.148, 1.000, 0.928, 0.767, 0.701, 0.807, 0.902,
                  0.997, 1.224, 1.089, 0.973, 0.980, 0.665])
    gas_fit_E = np.array([0.665, 0.949, 1.224])
    newdata = np.array([0.6650000, 0.7581667, 0.8513333, 0.9445000,
                        1.0376667, 1.1308333, 1.2240000])
    alpha = 0.01

    rfile = data_path / 'gas_result'
    results = []
    with open(rfile, 'r') as f:
        for i in range(8):
            f.readline()
            z = np.fromiter(
                (float(v) for v in f.readline().rstrip().split()),
                np.float_)
            results.append(z)
    return (E, NOx, gas_fit_E, newdata, alpha, results)


class TestLoess2d(object):
    "Test class for lowess."
    d = madeup_data()

    def test_2dbasic(self):
        "2D standard"
        (x, y, results, _, _) = self.d
        madeup = loess(x, y)
        madeup.model.span = 0.5
        madeup.model.normalize = True
        madeup.fit()
        npt.assert_almost_equal(madeup.outputs.fitted_values, results[0], 5)
        npt.assert_almost_equal(madeup.outputs.enp, 14.9, 1)
        npt.assert_almost_equal(madeup.outputs.residual_scale, 0.9693, 4)

    def test_2d_modflags(self):
        "2D - modification of model flags"
        (x, y, results, _, _) = self.d
        madeup = loess(x, y)
        madeup.model.span = 0.8
        madeup.model.drop_square = [True, False]
        madeup.model.parametric = [True, False]
        npt.assert_equal(madeup.model.parametric[:2], [1, 0])
        madeup.fit()
        npt.assert_almost_equal(madeup.outputs.fitted_values, results[1], 5)
        npt.assert_almost_equal(madeup.outputs.enp, 6.9, 1)
        npt.assert_almost_equal(madeup.outputs.residual_scale, 1.4804, 4)

    def test_2d_modfamily(self):
        "2D - family modification"
        (x, y, results, _, _) = self.d
        madeup = loess(x, y)
        madeup.model.span = 0.8
        madeup.model.drop_square = [True, False]
        madeup.model.parametric = [True, False]
        madeup.model.family = "symmetric"
        madeup.fit()
        npt.assert_almost_equal(madeup.outputs.fitted_values, results[2], 5)
        npt.assert_almost_equal(madeup.outputs.enp, 6.9, 1)
        npt.assert_almost_equal(madeup.outputs.residual_scale, 1.0868, 4)

    def test_2d_modnormalize(self):
        "2D - normalization modification"
        (x, y, results, _, _) = self.d
        madeup = loess(x, y)
        madeup.model.span = 0.8
        madeup.model.drop_square = [True, False]
        madeup.model.parametric = [True, False]
        madeup.model.family = "symmetric"
        madeup.model.normalize = False
        madeup.fit()
        npt.assert_almost_equal(madeup.outputs.fitted_values, results[3], 5)
        npt.assert_almost_equal(madeup.outputs.enp, 6.9, 1)
        npt.assert_almost_equal(madeup.outputs.residual_scale, 1.0868, 4)

    def test_2d_pred_nostderr(self):
        "2D prediction - no stderr"
        (x, y, results, newdata1, _) = self.d
        madeup = loess(x, y)
        madeup.model.span = 0.5
        madeup.model.normalize = True
        prediction = madeup.predict(newdata1, stderror=False)
        npt.assert_almost_equal(prediction.values, results[4], 5)
        #
        prediction = madeup.predict(newdata1, stderror=False)
        npt.assert_almost_equal(prediction.values, results[4], 5)

    def test_2d_pred_nodata(self):
        "2D prediction - nodata"
        (x, y, _, _, _) = self.d
        madeup = loess(x, y)
        try:
            madeup.predict(None)
        except ValueError:
            pass
        else:
            raise AssertionError("The test should have failed")

    def test_2d_pred_stderr(self):
        "2D prediction - w/ stderr"
        (x, y, results, _, newdata2) = self.d
        madeup = loess(x, y)
        madeup.model.span = 0.5
        madeup.model.normalize = True
        prediction = madeup.predict(newdata2, stderror=True)
        npt.assert_almost_equal(prediction.values, results[5], 5)
        npt.assert_almost_equal(prediction.stderr, [0.276746, 0.278009], 5)
        npt.assert_almost_equal(prediction.residual_scale, 0.969302, 6)
        npt.assert_almost_equal(prediction.df, 81.2319, 4)

        # Direct access
        prediction = madeup.predict(newdata2, stderror=True)
        npt.assert_almost_equal(prediction.values, results[5], 5)
        npt.assert_almost_equal(prediction.stderr, [0.276746, 0.278009], 5)
        npt.assert_almost_equal(prediction.residual_scale, 0.969302, 6)
        npt.assert_almost_equal(prediction.df, 81.2319, 4)

    def test_2d_pred_confinv(self):
        "2D prediction - confidence"
        (x, y, results, _, newdata2) = self.d
        madeup = loess(x, y)
        madeup.model.span = 0.5
        madeup.model.normalize = True
        prediction = madeup.predict(newdata2, stderror=True)
        ci = prediction.confidence(alpha=0.01)
        npt.assert_almost_equal(ci.lower, results[6][::3], 5)
        npt.assert_almost_equal(ci.fit, results[6][1::3], 5)
        npt.assert_almost_equal(ci.upper, results[6][2::3], 5)


class TestLoessGas(object):
    "Test class for lowess."

    d = gas_data()

    def test_1dbasic(self):
        "Basic test 1d"
        (E, NOx, _, _, _, results) = self.d
        gas = loess(E, NOx)
        gas.model.span = 2./3.
        gas.fit()
        npt.assert_almost_equal(gas.outputs.fitted_values, results[0], 6)
        npt.assert_almost_equal(gas.outputs.enp, 5.5, 1)
        npt.assert_almost_equal(gas.outputs.residual_scale, 0.3404, 4)

    def test_1dbasic_alt(self):
        "Basic test 1d - part #2"
        (E, NOx, _, _, _, results) = self.d
        gas_null = loess(E, NOx)
        gas_null.model.span = 1.0
        gas_null.fit()
        npt.assert_almost_equal(gas_null.outputs.fitted_values, results[1], 6)
        npt.assert_almost_equal(gas_null.outputs.enp, 3.5, 1)
        npt.assert_almost_equal(gas_null.outputs.residual_scale, 0.5197, 4)

    def test_1dpredict(self):
        "Basic test 1d - prediction"
        (E, NOx, gas_fit_E, _, _, results) = self.d
        gas = loess(E, NOx, span=2./3.)
        gas.fit()
        predicted = gas.predict(gas_fit_E, stderror=False).values
        npt.assert_almost_equal(predicted, results[2], 6)

    def test_1dpredict_2(self):
        "Basic test 1d - new predictions"
        (E, NOx, _, newdata, _, results) = self.d
        # gas = loess(E, NOx, span=2./3.)
        gas = loess(E, NOx)
        gas.model.span = 2./3.
        prediction = gas.predict(newdata, stderror=True)
        ci = prediction.confidence(alpha=0.01)
        npt.assert_almost_equal(ci.lower, results[3][0::3], 6)
        npt.assert_almost_equal(ci.fit, results[3][1::3], 6)
        npt.assert_almost_equal(ci.upper, results[3][2::3], 6)

    def test_anova(self):
        "Tests anova"
        (E, NOx, _, _, _, results) = self.d
        gas = loess(E, NOx, span=2./3.)
        gas.fit()
        gas_null = loess(E, NOx, span=1.0)
        gas_null.fit()
        gas_anova = loess_anova(gas, gas_null)
        gas_anova_theo = results[4]
        npt.assert_almost_equal(gas_anova.dfn, gas_anova_theo[0], 5)
        npt.assert_almost_equal(gas_anova.dfd, gas_anova_theo[1], 5)
        npt.assert_almost_equal(gas_anova.F_value, gas_anova_theo[2], 5)
        npt.assert_almost_equal(gas_anova.Pr_F, gas_anova_theo[3], 5)

    def test_failures(self):
        "Tests failures"
        (E, NOx, gas_fit_E, _, _, _) = self.d
        gas = loess(E, NOx, span=2./3.)
        # This one should fail (all parametric)
        gas.model.parametric = True
        with pytest.raises(ValueError):
            gas.fit()

        # This one also (all drop_square)
        gas.model.drop_square = True
        with pytest.raises(ValueError):
            gas.fit()

        gas.model.degree = 1
        with pytest.raises(ValueError):
            gas.fit()

        # This one should not (revert to std)
        gas.model.parametric = False
        gas.model.drop_square = False
        gas.model.degree = 2
        gas.fit()

        # Now, for predict .................
        prediction = gas.predict(gas_fit_E, stderror=False)
        # This one should fail (extrapolation & blending)
        with pytest.raises(ValueError):
            gas.predict(prediction.values, stderror=False)

        # But this one should not ..........
        gas.predict(gas_fit_E, stderror=False)

    def test_pickling(self):
        rs = np.random.RandomState(seed=123)
        n = 500
        np.arange(n)
        x = np.linspace(0, 500, n)
        y = rs.randint(1, 1000, n)
        lo = loess(y, x, span=0.3)
        lo.fit()

        # pickle lo, unpickle it as lo2 and compare
        # predictions from both
        with tempfile.TemporaryFile() as f:
            pickle.dump(lo, f)
            f.seek(0)
            lo2 = pickle.load(f)

        npt.assert_array_equal(lo.predict(x).values, lo2.predict(x).values)



"""
=================================================
Locally-weighted regression (:mod:`skmisc.loess`)
=================================================

Loess is a procedure for estimating a regression surface by a multivariate
smoothing procedure. A linear or quadratic function of the independent
variables is fit in a moving fashion that is analogous to how a moving
average is computed for a time series.

Compared to approaches that fit global parametric functions, loess
substantially increases the domain of surfaces that can be estimated without
distortion. However, analogues of the statistical procedures used
in parametric function fitting -- for example, ANOVA and t intervals --
involve statistics whose distributions are well approximated by familiar
distributions.

.. autosummary::
   :toctree: generated/
   :template: cython_class.rst

   loess
   loess_inputs
   loess_model
   loess_control
   loess_outputs
   loess_prediction
   loess_confidence_intervals
   loess_anova

Source
------
The original source code was written by William S. Cleveland, Eric Grosse
and Ming-Jen Shyu. It is available at http://www.netlib.org/a/dloess. It
was initially adapted to for use in Scipy by Pierre GF Gerard-Marchant.

For more see references [1]_ [2]_ and [3]_.

.. [1] W. S. Cleveland, E. Grosse, and M. J. Shyu. Local Regression Models.
   In J. M. Chambers and T. Hastie, editors, Statistical Models in S,
   pages 309--376. Chapman and Hall, New York, 1992.

.. [2] W. S. Cleveland, S. J. Devlin, and E. Grosse. Regression by Local
   Fitting: Methods, Properties, and Computing. Journal of Econometrics, 37:
   pp. 87--114. 1988.

.. [3] W. S. Cleveland. Robust Locally Weighted Regression and Smoothing
   Scatterplots. Journal of the American Statistical Association, 74:
   pp. 829--836. 1979.
"""
from ._loess import (
    loess,
    loess_anova,
    loess_confidence_intervals,
    loess_control,
    loess_inputs,
    loess_model,
    loess_outputs,
    loess_prediction,
)

__all__ = ['loess', 'loess_model', 'loess_control', 'loess_inputs',
           'loess_model', 'loess_outputs', 'loess_prediction',
           'loess_confidence_intervals', 'loess_anova']



from docutils.core import Publisher
from docutils.parsers.rst.roles import code_role


# Monkey patch docutils to give short css names
# credit: http://stackoverflow.com/a/21766874/832573
def process_programmatic_settings(self, settings_spec,
                                  settings_overrides,
                                  config_section):
    if self.settings is None:
        defaults = (settings_overrides or {}).copy()
        defaults.setdefault('traceback', True)
        defaults.setdefault('syntax_highlight', 'short')
        self.get_settings(settings_spec=settings_spec,
                          config_section=config_section,
                          **defaults)

Publisher.process_programmatic_settings = process_programmatic_settings


def python_role(role, rawtext, text, lineno,
                inliner, options={}, content=[]):
    options = {'language': 'python'}
    return code_role(role, rawtext, text, lineno,
                     inliner, options=options, content=content)


def setup(app):
    app.add_role('python', python_role)
    app.add_role('py', python_role)



import os
import shutil
import sys
from pathlib import Path

import click
from spin import util
from spin.cmds import meson


@click.command()
@click.option(
    "--build-dir",
    default="build",
    help="Build directory; default is `$PWD/build`"
)
@click.option(
    "--clean",
    is_flag=True,
    help="Clean previously built docs before building"
)
def docs(build_dir, clean=False):
    """
    📖 Build documentation
    """
    if clean:
        doc_dir = Path("./doc/build").absolute()
        if doc_dir.is_dir():
            print(f"Removing `{doc_dir}`")
            shutil.rmtree(doc_dir)

    site_path = meson._get_site_packages()
    if site_path is None:
        print("No built scikit-misc found; run `./spin build` first.")
        sys.exit(1)

    util.run(["pip", "install", "-q", "-r", "requirements/docs.txt"])

    PYTHONPATH = os.environ.get("PYTHONPATH", "")
    os.environ["SPHINXOPTS"] = "-W"
    os.environ['PYTHONPATH'] = f'{site_path}{os.sep}:{PYTHONPATH}'
    util.run(["make", "-C", "doc", "html"], replace=True)


@click.command()
@click.option(
    "--build-dir",
    default="build",
    help="Build directory; default is `$PWD/build`"
)
def coverage(build_dir):
    """
    📊 Generate coverage report
    """
    site_path = meson._get_site_packages()
    util.run([
        "python",
        "-m",
        "coverage",
        "report",
        "--data-file",
        f"{site_path}/.coverage",
    ], replace=True)


@click.command()
@click.option(
    "--build-dir",
    default="build",
    help="Build directory; default is `$PWD/build`"
)
def coverage_html(build_dir):
    """
    📊 Generate HTML coverage report
    """
    site_path = meson._get_site_packages()
    util.run([
        "python",
        "-m",
        "coverage",
        "html",
        "--data-file",
        f"{site_path}/.coverage",
    ], replace=True)


@click.command()
def sdist():
    """
    📦 Build a source distribution in `build/meson-dist/`.
    """
    # Using the build module gives better results than using
    # meson directory. It creates an sdist with PKG-INFO
    util.run([
        "python",
        "-m",
        "build",
        "--no-isolation",
        "--skip-dependency-check",
        "--sdist",
        ".",
    ], replace=True)



# -*- Mode: Python -*-
# cython: embedsignature=True
# cython: language_level=3

# TODO: When Cython3 is release
# Change from the deprecated the numpy 1.7 API, see
#    - https://github.com/cython/cython/issues/2498
#    - https://github.com/scipy/scipy/pull/10840#issuecomment-549939230

import cython
import numpy as np
cimport numpy as np
from numpy cimport (
    ndarray,
    npy_intp,
    NPY_DOUBLE,
    PyArray_SimpleNewFromData,
)
cimport c_loess   # c_loess.pxd

# NumPy must be initialized
np.import_array()


cdef floatarray_from_data(int rows, int cols, double *data):
    cdef ndarray a_ndr
    cdef npy_intp dims[2]
    cdef int nd
    dims = (rows, cols)
    nd = 2 if cols > 1 else 1
    a_ndr = <object>PyArray_SimpleNewFromData(nd, dims, NPY_DOUBLE, data)
    # Since the memory is owned by the containing object, we pass a copy
    # of the data so that when the owner is destroyed the return data
    # is not destroyed.
    return a_ndr.copy()


cdef boolarray_from_data(int rows, int cols, int *data):
    cdef ndarray a_ndr
    cdef npy_intp dims[2]
    cdef int nd
    dims = (rows, cols)
    nd = 2 if cols > 1 else 1
    a_ndr = <object>PyArray_SimpleNewFromData(nd, dims, NPY_DOUBLE, data)
    # Since the memory is owned by the containing object, we pass a copy
    # of the data so that when the owner is destroyed the return data
    # is not destroyed.
    return a_ndr.astype(bool).copy()


cdef class loess_inputs:
    """
    Initialization class for loess data inputs

    Parameters
    ----------
    x : ndarray[n, p]
        n independent observations for p no. of variables
    y : ndarray[n]
        A (n,) ndarray of response observations
    weights : ndarray[n] or None
        Weights to be given to individual observations
        in the sum of squared residuals that forms the local fitting
        criterion. If not None, the weights should be non negative. If
        the different observations have non-equal variances, the weights
        should be inversely proportional to the variances. By default,
        an unweighted fit is carried out (all the weights are one).
    """
    cdef c_loess.c_loess_inputs _base
    cdef readonly allocated

    def __cinit__(self, x, y, weights=None):
        cdef ndarray _x, _y, _w
        # When errors are raised the object gets destroyed and
        # __dealloc__ is called. It should not free up memory
        # if none was allocated prior to the error.
        self.allocated = False

        x = np.array(x, copy=False, subok=True,
                      dtype=np.float_, order='C')
        y = np.array(y, copy=False, subok=True,
                      dtype=np.float_, order='C')
        n = len(x)

        # Check the dimensions
        if x.ndim == 1:
            p = 1
        elif x.ndim == 2:
            p = x.shape[1]
        else:
            raise ValueError("The array of indepedent varibales "
                             "should be 2D at most!")

        if y.ndim != 1:
            raise ValueError("The array of dependent variables "
                             "should be 1D.")
        elif n != len(y):
            raise ValueError("The independent and depedent varibales "
                             "should have the same number of "
                             "observations.")

        if weights is None:
            weights = np.ones((n,), dtype=np.float_)

        if weights.ndim > 1 or weights.size != n:
            raise ValueError("Invalid size of the 'weights' vector!")

        weights = np.array(weights, copy=False, subok=True,
                           dtype=np.float_, order='C')

        # Python objects -> C structures -> *data in C structures
        _x = x.ravel()
        _y = y
        _w = weights

        x_dat = <double *>_x.data
        y_dat = <double *>_y.data
        w_dat = <double *>_w.data

        c_loess.loess_inputs_setup(x_dat, y_dat, w_dat, n, p, &self._base)
        self.allocated = True

    def __init__(self, x, y, weights=None):
        # For documentation
        pass

    def __dealloc__(self):
        if self.allocated:
            c_loess.loess_inputs_free(&self._base)

    @property
    def n(self):
        """
        :class:`int` - Number of independent observations
        """
        return self._base.n

    @property
    def p(self):
        """
        :class:`int` - Number of variables
        """
        return self._base.p

    @property
    def x(self):
        """
        :class:`~numpy.ndarray` - Independent observations, shape ``(n, p)``
        """
        return floatarray_from_data(self.n, self.p, self._base.x)

    @property
    def y(self):
        """
        :class:`~numpy.ndarray` - Response observations, shape ``(n,)``
        """
        return floatarray_from_data(self.n, 1, self._base.y)


cdef class loess_control:
    """
    Initialization class for loess control parameters

    Parameters
    ----------
    surface : str, optional
        One of ['interpolate', 'direct']
        Determines whether the fitted surface is computed directly
        at all points ('direct') or whether an interpolation method
        is used ('interpolate'). The default 'interpolate') is what
        most users should use unless special circumstances warrant.
    statistics : str, optional
        One of ['approximate', 'exact']
        Determines whether the statistical quantities are computed
        exactly ('exact') or approximately ('approximate'). 'exact'
        should only be used for testing the approximation in
        statistical development and is not meant for routine usage
        because computation time can be horrendous.
    trace_hat : str, optional
        One of ['wait.to.decide', 'exact', 'approximate']
        Determines how the trace of the hat matrix should be computed.
        The hat matrix is used in the computation of the statistical
        quantities. If 'exact', an exact computation is done; this
        could be slow when the number of observations n becomes large.
        If 'wait.to.decide' is selected, then a default is 'exact'
        for n < 500 and 'approximate' otherwise.
        This option is only useful when the fitted surface is
        interpolated. If surface is 'exact', an exact computation is
        always done for the trace. Setting trace_hat to 'approximate'
        for large dataset will substantially reduce the computation time.
    iterations : int, optional
        Number of iterations of the robust fitting method. If the family
        is 'gaussian', the number of iterations is set to 0.
    cell : float, optional
        Maximum cell size of the kd-tree. Suppose k = floor(n*cell*span),
        where n is the number of observations, and span the smoothing
        parameter. Then, a cell is further divided if the number of
        observations within it is greater than or equal to k. This
        option is only used if the surface is interpolated.
    """
    cdef c_loess.c_loess_control _base
    cdef bytes _surface
    cdef bytes _statistics
    cdef bytes _trace_hat

    def __cinit__(self, *args, **kwargs):
        # Does not allocate memory
        c_loess.loess_control_setup(&self._base)

    def __init__(self, surface='interpolate', statistics='approximate',
                 trace_hat='wait.to.decide', iterations=4, cell=0.2):
        self.surface = surface
        self.statistics = statistics
        self.trace_hat = trace_hat
        self.iterations = iterations
        self.cell = cell

    @property
    def surface(self):
        return self._base.surface.decode('utf-8')

    @surface.setter
    def surface(self, value):
        if value not in ('interpolate', 'direct'):
            raise ValueError(
                "Invalid value for the 'surface' argument: "
                "should be in ('interpolate', 'direct').")
        self._surface = value.encode('utf-8')
        self._base.surface = self._surface

    @property
    def statistics(self):
        return self._base.statistics.decode('utf-8')

    @statistics.setter
    def statistics(self, value):
        if value not in ('approximate', 'exact'):
            raise ValueError(
                "Invalid value for the 'statistics' argument: "
                "should be in ('approximate', 'exact').")
        self._statistics = value.encode('utf-8')
        self._base.statistics = self._statistics

    @property
    def trace_hat(self):
        return self._base.trace_hat.decode('utf-8')

    @trace_hat.setter
    def trace_hat(self, value):
        if value not in ('wait.to.decide', 'approximate', 'exact'):
            raise ValueError(
                "Invalid value for the 'trace_hat' argument: "
                "should be in ('approximate', 'exact').")
        self._trace_hat = value.encode('utf-8')
        self._base.trace_hat = self._trace_hat

    @property
    def iterations(self):
        return self._base.iterations

    @iterations.setter
    def iterations(self, value):
        if value < 0:
            raise ValueError(
                "Invalid number of iterations: "
                "should be positive")
        self._base.iterations = value

    @property
    def cell(self):
        return self._base.cell

    @cell.setter
    def cell(self, value):
        if value <= 0:
            raise ValueError(
                "Invalid value for the cell argument: "
                " should be positive")
        self._base.cell = value

    def __str__(self):
        strg = ["Control",
                "-------",
                "Surface type     : %s" % self.surface,
                "Statistics       : %s" % self.statistics,
                "Trace estimation : %s" % self.trace_hat,
                "Cell size        : %s" % self.cell,
                "Nb iterations    : %s" % self.iterations,]
        return '\n'.join(strg)


cdef class loess_kd_tree:
    cdef c_loess.c_loess_kd_tree _base

    def __cinit__(self, n, p):
        c_loess.loess_kd_tree_setup(n, p, &self._base)

    def __dealloc__(self):
        c_loess.loess_kd_tree_free(&self._base)


cdef class loess_model:
    """
    Initialization class for loess fitting parameters

    Parameters
    ----------
    p : int
        Number of variables
    family : str
        One of ('gaussian', 'symmetric')
        Determines the assumed distribution of the errors. If 'gaussian'
        the fit is performed with least-squares. If 'symmetric' is
        selected, the fit is performed robustly by redescending
        M-estimators.
    span : float
        Smoothing factor, as a fraction of the number of points to take
        into account. Should be in the range (0, 1]. Default is 0.75
    degree : int
        Overall degree of locally-fitted polynomial. 1 is locally-linear
        fitting and 2 is locally-quadratic fitting. Degree should be 2 at
        most. Default is 2.
    normalize : bool
        Determines whether the independent variables should be normalized.
        If True, the normalization is performed by setting the 10% trimmed
        standard deviation to one. If False, no normalization is carried
        out. This option is only useful for more than one variable. For
        spatial coordinates predictors or variables with a common scale,
        it should be set to False. Default is True.
    parametric : bool | list-of-bools of length p
        Indicates which independent variables should be
        conditionally-parametric (if there are two or more independent
        variables). If a sequence is given, the values should be ordered
        according to the predictor group in x.
    drop_square : bool | list-of-bools of length p
        Which squares to drop. When there are two or more independent
        variables and when a 2nd order polynomial(degree) is used,
        'drop_square' specifies those numeric predictors
        whose squares should be dropped from the set of fitting variables.
        If a sequence is given, the values should be ordered according to
        the predictor group in x.
    """

    cdef c_loess.c_loess_model _base
    cdef p
    cdef bytes _family

    def __cinit__(self, *args, **kwargs):
        # Does not allocate memory
        c_loess.loess_model_setup(&self._base)

    def __init__(self, p, family='gaussian', span=0.75,
                 degree=2, normalize=True, parametric=False,
                 drop_square=False):
        self.p = p
        self.family = family
        self.span = span
        self.degree = degree
        self.normalize = normalize
        self.parametric = parametric
        self.drop_square = drop_square

    @property
    def normalize(self):
        return bool(self._base.normalize)

    @normalize.setter
    def normalize(self, value):
        self._base.normalize = value

    @property
    def span(self):
        return self._base.span

    @span.setter
    def span(self, value):
        if value <= 0. or value > 1.:
            raise ValueError("Span should be between 0 and 1!")
        self._base.span = value

    @property
    def degree(self):
        return self._base.degree

    @degree.setter
    def degree(self, value):
        if value < 0 or value > 2:
            raise ValueError("Degree should be be 0, 1 or 2!")
        self._base.degree = value

    @property
    def family(self):
        return self._base.family.decode('utf-8')

    @family.setter
    def family(self, value):
        if value.lower()  not in ('symmetric', 'gaussian'):
            raise ValueError(
                "Invalid value for the 'family' argument: "
                "should be in ('symmetric', 'gaussian').")
        self._family = value.encode('utf-8')
        self._base.family = self._family

    @property
    def parametric(self):
        return boolarray_from_data(self.p, 1, self._base.parametric)

    @parametric.setter
    def parametric(self, value):
        cdef ndarray p_ndr

        if value in (True, False):
            value = [value] * self.p
        elif len(value) != self.p:
            raise ValueError(
                "'parametric' should be a boolean or a list "
                "of booleans with length equal to the number "
                "of independent variables")

        p_ndr = np.atleast_1d(np.array(value, copy=False, subok=True,
                                       dtype=bool))
        for i in range(self.p):
            self._base.parametric[i] = p_ndr[i]

    @property
    def drop_square(self):
        return boolarray_from_data(self.p, 1, self._base.drop_square)

    @drop_square.setter
    def drop_square(self, value):
        cdef ndarray d_ndr

        if value in (True, False):
            value = [value] * self.p
        elif len(value) != self.p:
            raise ValueError(
                "'drop_square' should be a boolean or a list "
                "of booleans with length equal to the number "
                "of independent variables")

        d_ndr = np.atleast_1d(np.array(value, copy=False,
                                       subok=True, dtype=bool))
        for i in range(self.p):
            self._base.drop_square[i] = d_ndr[i]

    def __repr__(self):
        return "<loess object: model parameters>"

    def __str__(self):
        strg = ["Model parameters",
                "----------------",
                "Family          : %s" % self.family,
                "Span            : %s" % self.span,
                "Degree          : %s" % self.degree,
                "Normalized      : %s" % self.normalize,
                "Parametric      : %s" % self.parametric[:self.p],
                "Drop_square     : %s" % self.drop_square[:self.p],
                ]
        return '\n'.join(strg)


cdef class loess_outputs:
    """
    Class of a loess fit outputs

    This object is automatically created with empty values when a
    new loess object is instantiated. The object gets filled when the
    loess.fit() method is called.

    Parameters
    ----------
    n : int
        Number of independent observation
    p : int
        Number of variables
    """
    cdef c_loess.c_loess_outputs _base
    # private
    cdef readonly family
    cdef readonly n, p
    cdef readonly activated

    def __cinit__(self, n, p, family):
        c_loess.loess_outputs_setup(n, p, &self._base)

    def __init__(self, n, p, family):
        self.n = n
        self.p = p
        self.family = family
        self.activated = False

    def __dealloc__(self):
        c_loess.loess_outputs_free(&self._base)

    @property
    def fitted_values(self):
        """
        :class:`~numpy.ndarray` - Fitted values, shape ``(n,)``
        """
        return floatarray_from_data(self.n, 1, self._base.fitted_values)

    @property
    def fitted_residuals(self):
        """
        :class:`~numpy.ndarray` - Fitted residuals, shape ``(n,)``,
        (observations - fitted values)
        """
        return floatarray_from_data(self.n, 1, self._base.fitted_residuals)

    @property
    def pseudovalues(self):
        """
        :class:`~numpy.ndarray` - Adjusted values of the response
        when robust estimation is used, shape ``(n,)``
        """
        if self.family != 'symmetric':
            raise ValueError(
                "pseudovalues are available only when "
                "robust fitting. Use family='symmetric' "
                "for robust fitting")
        return floatarray_from_data(self.n, 1, self._base.pseudovalues)

    @property
    def diagonal(self):
        """
        :class:`~numpy.ndarray` - Diagonal of the operator hat matrix,
        shape ``(n,)``
        """
        return floatarray_from_data(self.n, 1, self._base.diagonal)

    @property
    def robust(self):
        """
        :class:`~numpy.ndarray` - Robustness weights for robust fitting,
        shape ``(n,)``
        """
        return floatarray_from_data(self.n, 1, self._base.robust)

    @property
    def divisor(self):
        """
        :class:`~numpy.ndarray` - Normalization divisors for numeric
        predictors, shape ``(p,)``
        """
        return floatarray_from_data(self.n, 1, self._base.divisor)

    @property
    def enp(self):
        """
        :class:`float` - Equivalent number of parameters
        """
        return self._base.enp

    @property
    def residual_scale(self):
        """
        :class:`float` - Estimate of the scale of residuals
        """
        return self._base.residual_scale

    @property
    def one_delta(self):
        """
        :class:`float` - Statistical parameter used in the computation
        of standard errors
        """
        return self._base.one_delta

    @property
    def two_delta(self):
        """
        :class:`float` - Statistical parameter used in the computation
        of standard errors
        """
        return self._base.two_delta

    @property
    def trace_hat(self):
        """
        :class:`float` - Trace of the operator hat matrix
        """
        return self._base.trace_hat

    def __str__(self):
        strg = ["Outputs",
                "-------",
                "Fitted values         : %s\n" % self.fitted_values,
                "Fitted residuals      : %s\n" % self.fitted_residuals,
                "Eqv. nb of parameters : %s" % self.enp,
                "Residual Scale        : %s" % self.residual_scale,
                "Deltas                : %s - %s" % (self.one_delta,
                                                     self.two_delta),
                "Normalization factors : %s" % self.divisor,]
        return '\n'.join(strg)


cdef class loess_confidence_intervals:
    """
    Pointwise confidence intervals of a loess-predicted object

    Parameters
    ----------
    pred : loess_prediction
        Prediction object
    alpha : float
        The alpha level for the confidence interval.
        It must be in the range (0, 1)
    """
    cdef c_loess.c_confidence_intervals _base
    cdef readonly m

    def __cinit__(loess_confidence_intervals self, loess_prediction pred,
                  float alpha):
        coverage = 1 - alpha
        if coverage < .5:
            coverage = 1 - coverage

        if not 0 < alpha < 1. :
            raise ValueError("The alpha value should be "
                             "between 0 and 1.")
        if not pred._base.se:
            raise ValueError("Cannot compute confidence intervals "
                             "without standard errors.")

        c_loess.c_pointwise(&pred._base, coverage, &self._base)
        self.m = pred.m

    def __init__(self, pred, alpha):
        # For documentation
        pass

    def __dealloc__(self):
        c_loess.pw_free_mem(&self._base)

    @property
    def fit(self):
        """
        :class:`~numpy.ndarray` - Predicted values
        """
        return floatarray_from_data(self.m, 1, self._base.fit)

    @property
    def upper(self):
        """
        :class:`~numpy.ndarray` - Upper bounds of the confidence intervals
        """
        return floatarray_from_data(self.m, 1, self._base.upper)

    @property
    def lower(self):
        """
        :class:`~numpy.ndarray` - Lower bounds of the confidence intervals
        """
        return floatarray_from_data(self.m, 1, self._base.lower)


cdef class loess_prediction:
    """
    Class for loess prediction results

    Holds the predicted values and standard errors of a loess object

    Parameters
    ----------
    newdata : ndarray[m, p]
        Independent variables where the surface must be estimated,
        with m the number of new data points, and p the number of
        independent variables.
    loess : loess.loess
        Loess object that has been successfully fitted,
        i.e `loess.fit` has been called and it returned without
        any errors.
    stderror : boolean
        Whether the standard error should be computed
    """

    cdef c_loess.c_prediction _base
    cdef readonly allocated

    def __cinit__(self, newdata, loess loess, stderror=False):
        cdef ndarray p_ndr
        cdef double *p_dat
        cdef int se

        self.allocated = False

        # Note : we need a copy as we may have to normalize
        p_ndr = np.array(newdata, copy=True, subok=True, order='C')
        p_ndr = p_ndr.astype(float)

        # Dimensions should match those of the input
        if p_ndr.size == 0 or p_ndr.ndim == 0:
            raise ValueError("Can't predict without input data !")

        if p_ndr.ndim > 2:
            raise ValueError("New data has more than 2 dimensions.")

        _p = 1 if p_ndr.ndim == 1 else p_ndr.shape[1]
        if _p != loess.inputs.p:
            msg = ("Incompatible data size: there should be as many "
                   "columns as parameters. Got %d instead of %d "
                   "parameters" % (_p, loess.inputs.p))
            raise ValueError(msg)

        se = 1 if stderror else 0
        m = len(p_ndr)
        p_ndr = p_ndr.ravel()
        p_dat = <double *>p_ndr.data

        c_loess.predict_setup(&self._base, &loess._base, se, m)
        self.allocated = True
        c_loess.c_predict(p_dat, &loess._base, &self._base)

        if loess._base.status.err_status:
            raise ValueError(loess._base.status.err_msg)

    def __init__(self, newdata, loess, stderror=False):
        # For documentation
        pass

    def __dealloc__(self):
        if self.allocated:
            c_loess.predict_free(&self._base)

    @property
    def values(self):
        """
        :class:`~numpy.ndarray` - loess values evaluated at newdata,
        shape ``(m,)``
        """
        return floatarray_from_data(self.m, 1, self._base.fit)

    @property
    def stderr(self):
        """
        :class:`~numpy.ndarray` - Estimates of the standard error on
        the estimated values, shape ``(m,)``

        Raises `ValueError` if the standard error was not computed.
        """
        if not self._base.se:
            raise ValueError("Standard error was not computed."
                             "Use 'stderror=True' when predicting.")
        return floatarray_from_data(self.m, 1, self._base.se_fit)

    @property
    def residual_scale(self):
        """
        :class:`float` - Estimate of the scale of the residuals
        """
        return self._base.residual_scale

    @property
    def df(self):
        """
        :class:`float` - Degrees of freedom of the loess fit.

        It is used with the t-distribution to compute pointwise
        confidence intervals for the evaluated surface. It is
        obtained using the formula ``(one_delta ** 2) / two_delta``.
        """
        return self._base.df

    @property
    def m(self):
        """
        :class:`int` - Number of observations in the new data points
        """
        return self._base.m

    def confidence(self, alpha=0.05):
        """
        Returns the pointwise confidence intervals

        Parameters
        ----------
        alpha : float
            The alpha level for the confidence interval. The
            default ``alpha=0.05`` returns a 95% confidence
            interval. Therefore it must be in the range (0, 1).

        Returns
        -------
        out : loess_confidence_intervals
            Confidence intervals object. It has attributes `fit`,
            `lower` and `upper`
        """
        return loess_confidence_intervals(self, alpha)

    def __str__(self):
        try:
            stderr = "Predicted std error   : %s\n" % self.stderr
        except ValueError:
            stderr = ""

        strg = ["Outputs",
                "-------",
                "Predicted values      : %s\n" % self.values,
                stderr,
                "Residual scale        : %s" % self.residual_scale,
                "Degrees of freedom    : %s" % self.df,
                ]
        return '\n'.join(strg)

def _new_loess(init_arguments, fit):
    """
    Create the loess object from initial arguments

    Used for pickling
    """
    # Generate the object from the initialising arguments
    # Run the fit method if it was run before
    l = loess(**init_arguments)
    if fit:
        l.fit()
    return l


cdef class loess:
    """
    Locally-weighted regression

    A loess object is initialized with the combined parameters of
    :class:`loess_inputs`, :class:`loess_model` and
    :class:`loess_control`. The parameters of :class:`loess_inputs`
    i.e ``x``, ``y`` and ``weights`` can be positional in that order.
    In the descriptions below, `n` is the number of observations,
    and `p` is the number of predictor variables.

    Parameters
    ----------
    x : ndarray[n, p]
        n independent observations for p no. of variables
    y : ndarray[n,]
        A (n,) ndarray of response observations
    weights : ndarray[n] or None
        Weights to be given to individual observations
        in the sum of squared residuals that forms the local fitting
        criterion. If not None, the weights should be non negative. If
        the different observations have non-equal variances, the weights
        should be inversely proportional to the variances. By default,
        an unweighted fit is carried out (all the weights are one).
    **options : dict
        The parameters of :class:`loess_model` and
        :class:`loess_control`.

    Attributes
    ----------
    inputs : :class:`loess_inputs`
        Object that handles the inputs
    model : :class:`loess_model`
        Object that handles the model
    control : :class:`loess_control`
        Object that holds the control parameters
    kd_tree : :class:`loess_kdtree`
        Object that holds the parameters and structures used
        internally by the regression algorithm.
    outputs : :class:`loess_outputs`
        Object that holds the output values and parameters.
        These should be read after :meth:`loess.fit` has been
        called.

    Note
    ----
    Loess smoothing creates large state, so pickling a fitted loess
    object does not save the state. It saves the `__init__` parameters
    from which the state is recreated (refitting) when unpickled. So
    pickling and unpickling does not save on computation time.
    """
    cdef c_loess.c_loess _base
    cdef readonly loess_inputs inputs
    cdef readonly loess_model model
    cdef readonly loess_control control
    cdef readonly loess_kd_tree kd_tree
    cdef readonly loess_outputs outputs

    # save all the arguments to enable pickling
    cdef object _init_arguments

    def __init__(self, object x, object y, object weights=None, **options):
        self._init_arguments = {
            "x": x, "y": y, "weights": weights, **options
        }

        # Process options
        model_options = {}
        control_options= {}
        for (k, v) in options.items():
            if k in ('family', 'span', 'degree', 'normalize',
                     'parametric', 'drop_square',):
                model_options[k] = v
            elif k in ('surface', 'statistics', 'trace_hat',
                       'iterations', 'cell'):
                control_options[k] = v

        # Initialize the inputs
        self.inputs = loess_inputs(x, y, weights)
        self._base.inputs = &self.inputs._base

        n = self.inputs.n
        p = self.inputs.p

        # Initialize the control parameters
        self.control = loess_control(**control_options)
        self._base.control = &self.control._base

        # Initialize the model parameters
        self.model = loess_model(p, **model_options)
        self._base.model = &self.model._base

        # Initialize the outputs
        self.outputs = loess_outputs(n, p, self.model.family)
        self._base.outputs = &self.outputs._base

        # Initialize the kd tree
        self.kd_tree = loess_kd_tree(n, p)
        self._base.kd_tree = &self.kd_tree._base

    def fit(self):
        """
        Computes the loess parameters on the current inputs and
        sets of parameters.
        """
        c_loess.loess_fit(&self._base)
        self.outputs.activated = True
        if self._base.status.err_status:
            raise ValueError(self._base.status.err_msg)
        return

    def input_summary(self):
        """
        Returns some generic information about the loess parameters.
        """
        toprint = [str(self.model), str(self.control)]
        return "\n\n".join(toprint)

    def output_summary(self):
        """Returns some generic information about the loess fit."""
        fit_flag = bool(self.outputs.activated)

        if self.model.family == "gaussian":
            rse = ("Residual Standard Error        : %.4f" %
                   self.outputs.residual_scale)
        else:
            rse = ("Residual Scale Estimate        : %.4f" %
                   self.outputs.residual_scale)

        strg = ["Output Summary",
                "--------------",
                "Number of Observations         : %d" % self.inputs.n,
                "Fit flag                       : %d" % fit_flag,
                "Equivalent Number of Parameters: %.1f" % self.outputs.enp,
                rse
                ]
        return '\n'.join(strg)

    def predict(self, newdata, stderror=False):
        """
        Compute loess estimates at the given new data points newdata.

        Parameters
        ----------
        newdata : ndarray[m, p]
            Independent variables where the surface must be estimated,
            with m the number of new data points, and p the number of
            independent variables.
        stderror : boolean
            Whether the standard error should be computed

        Returns
        -------
        A :class:`loess_prediction` object.
        """
        # Make sure there's been a fit earlier
        if self.outputs.activated == 0:
            self.fit()

        return loess_prediction(newdata, self, stderror)

    # Pickling support
    @cython.final
    def __reduce__(self):
        return (_new_loess, (self._init_arguments, self.outputs.activated))

cdef class loess_anova:
    """
    Analysis of variance for two loess objects

    Parameters
    ----------
    loess_one : loess.loess
        First loess object
    loess_two : loess.loess
        Second loess object

    Attributes
    ----------
    F_value : float
        Value of the F-statistic
    Pr_F : float
        Probability of getting a value as large as the
        `F_value`. The is the p-value of the F-statistic.
    """
    cdef readonly double dfn, dfd, F_value, Pr_F

    def __init__(self, loess_one, loess_two):
        cdef double one_d1, one_d2, one_s, two_d1, two_d2, two_s
        cdef double rssdiff, d1diff, tmp, df1, df2

        if (not isinstance(loess_one, loess) or
                not isinstance(loess_two, loess)):
            raise ValueError("Arguments should be valid loess objects!"
                             "got '%s' instead" % type(loess_one))

        out_one = loess_one.outputs
        out_two = loess_two.outputs

        one_d1 = out_one.one_delta
        one_d2 = out_one.two_delta
        one_s = out_one.residual_scale

        two_d1 = out_two.one_delta
        two_d2 = out_two.two_delta
        two_s = out_two.residual_scale

        rssdiff = abs(one_s * one_s * one_d1 - two_s * two_s * two_d1)
        d1diff = abs(one_d1 - two_d1)
        self.dfn = d1diff * d1diff / abs(one_d2 - two_d2)
        df1 = self.dfn

        if out_one.enp > out_two.enp:
            self.dfd = one_d1 * one_d1 / one_d2
            tmp = one_s
        else:
            self.dfd = two_d1 * two_d1 / two_d2
            tmp = two_s
        df2 = self.dfd
        self.F_value = (rssdiff / d1diff) / (tmp * tmp)
        self.Pr_F = 1. - c_loess.pf(self.F_value, df1, df2)



#!/usr/bin/env python
import os
import shutil
from pathlib import Path

dist_root: str = os.getenv("MESON_DIST_ROOT")  # type: ignore
build_root: str = os.getenv("MESON_BUILD_ROOT")  # type: ignore


def copy_to_dist(rel_path: str) -> None:
    """
    Copy file at rel_path to the sdist directory
    """
    src = Path(build_root) / rel_path
    dest = Path(dist_root) / rel_path
    if src.exists():
        shutil.copy(src, dest)


if __name__ == "__main__":
    copy_to_dist("skmisc/_version.py")
